import math
from .core import Scalar, Edge


def _sigmoid(x):
    return 1 / (1 + math.exp(-x))


def _sigmoid_prime(x):
    return _sigmoid(x) * (1 - _sigmoid(x))


def _relu(x):
    return max([0, x])


def _relu_prime(x):
    if x < 0:
        return 0
    else:
        return 1


def _log_prime(x):
    if x != 0:
        return 1 / math.fabs(x)
    else:
        return 0


def _tanh_prime(x):
    return 1 - math.tanh(x) ** 2


def sin(x: Scalar):
    return Scalar(value=math.sin(x.value), prevs=[Edge(x, math.cos(x.value))])


def cos(x: Scalar):
    return Scalar(value=math.cos(x.value), prevs=[Edge(x, -math.sin(x.value))])


def exp(x: Scalar):
    return Scalar(value=math.exp(x.value), prevs=[Edge(x, math.exp(x.value))])


def log(x: Scalar):
    return Scalar(value=math.log(x.value), prevs=[Edge(x, _log_prime(x.value))])


def sigmoid(x: Scalar):
    return Scalar(
        value=_sigmoid(x.value),
        prevs=[Edge(x, _sigmoid_prime(x.value))],
    )


def relu(x: Scalar):
    return Scalar(value=_relu(x.value), prevs=[Edge(x, _relu_prime(x.value))])


def tanh(x: Scalar):
    return Scalar(value=math.tanh(x.value), prevs=[Edge(x, _tanh_prime(x.value))])



from typing import List, Callable


class GradientDescent:
    def __init__(self, model, loss: Callable):
        self.model = model
        self.loss = loss

    def run(self, xs, ys, lr: float = 0.01, n_steps: int = 1000):
        for _ in range(n_steps):
            preds = [self.model.forward(x) for x in xs]
            l = self.loss(preds, ys)
            l.backward()
            self.model.gradient_update(lr)



import numpy as np
from .core import Tensor, Edge
from .core.utils import _pointwise


import numpy as np


def _sigmoid(x: np.ndarray):
    return 1 / (1 + np.exp(-x))


def _sigmoid_prime(x: np.ndarray):
    sigmoid_x = _sigmoid(x)
    return sigmoid_x * (1 - sigmoid_x)


def _relu(x: np.ndarray):
    return x * (x > 0)


def _relu_prime(x: np.ndarray):
    return (x > 0).astype(int)


def _tanh_prime(x: np.ndarray):
    return 1 - np.tanh(x) ** 2


def exp(x: Tensor):
    return Tensor(
        value=np.exp(x.value),
        prevs=[Edge(prev=x, local_grad=np.exp(x.value), backward_fn=_pointwise)],
    )


def log(x: Tensor):
    return Tensor(
        value=np.log(x.value),
        prevs=[Edge(prev=x, local_grad=1 / np.abs(x.value), backward_fn=_pointwise)],
    )


def sin(x: Tensor):
    return Tensor(
        value=np.sin(x.value),
        prevs=[Edge(prev=x, local_grad=np.cos(x.value), backward_fn=_pointwise)],
    )


def cos(x: Tensor):
    return Tensor(
        value=np.cos(x.value),
        prevs=[Edge(prev=x, local_grad=-np.sin(x.value), backward_fn=_pointwise)],
    )


def sigmoid(x: Tensor):
    return Tensor(
        value=_sigmoid(x.value),
        prevs=[
            Edge(prev=x, local_grad=_sigmoid_prime(x.value), backward_fn=_pointwise)
        ],
    )


def relu(x: Tensor):
    return Tensor(
        value=_relu(x.value),
        prevs=[Edge(prev=x, local_grad=_relu_prime(x.value), backward_fn=_pointwise)],
    )


def tanh(x: Tensor):
    return Tensor(
        value=np.tanh(x.value),
        prevs=[Edge(prev=x, local_grad=_tanh_prime(x.value), backward_fn=_pointwise)],
    )



import math
import random
from collections import namedtuple


Edge = namedtuple("Edge", ["prev", "local_grad"])


class Scalar:
    """
    A scalar node of the computational graph.

    Attributes:
        value: The scalar value of the node.
        prevs: A list of Edge instances.
    """

    def __init__(self, value: float, prevs=None):
        self.value = value
        self.prevs = prevs if prevs is not None else []
        self.backwards_grad = 0

    def __repr__(self):
        return f"Scalar({self.value})"

    def _backward_step(self):
        for prev, local_grad in self.prevs:
            prev.backwards_grad += local_grad * self.backwards_grad

    def _get_graph(self, zero_grad=False):
        """
        Returns the Scalar nodes as a topologically ordered list, ending with self.

        Args:
            zero_grad: Resets the backwards_grad attribute for all nodes if true.

        Source: https://github.com/karpathy/micrograd/blob/master/micrograd/engine.py#L54
        """
        ordered_scalars = []
        visited_scalars = set()

        def traverse_graph(x):
            if x not in visited_scalars:
                visited_scalars.add(x)

                if zero_grad:
                    x.backwards_grad = 0

                for prev, _ in x.prevs:
                    traverse_graph(prev)

                ordered_scalars.append(x)

        traverse_graph(self)

        return ordered_scalars

    def _zero_grad(self):
        self._get_graph(zero_grad=True)

    def gradient_update(self, lr):
        self.value -= lr * self.backwards_grad

    def backward(self):
        ordered_scalars = self._get_graph(zero_grad=True)

        self.backwards_grad = 1

        for scalar in reversed(ordered_scalars):
            scalar._backward_step()

    def __add__(self, other):
        if not isinstance(other, Scalar):
            other = Scalar(other)

        return Scalar(
            value=self.value + other.value,
            prevs=[Edge(self, 1), Edge(other, 1)],
        )

    def __radd__(self, other):
        return self.__add__(other)

    def __sub__(self, other):
        return self.__add__(-other)

    def __rsub__(self, other):
        return other + self.__neg__()

    def __mul__(self, other):
        if not isinstance(other, Scalar):
            other = Scalar(other)

        return Scalar(
            value=self.value * other.value,
            prevs=[Edge(self, other.value), Edge(other, self.value)],
        )

    def __rmul__(self, other):
        return self.__mul__(other)

    def __neg__(self):
        return (-1) * self

    def __truediv__(self, other):
        if not isinstance(other, Scalar):
            other = Scalar(other)

        return Scalar(
            value=self.value / other.value,
            prevs=[
                Edge(self, 1 / other.value),
                Edge(other, -self.value / other.value**2),
            ],
        )

    def __rtruediv__(self, other):
        if not isinstance(other, Scalar):
            other = Scalar(other)

        return other / self

    def __pow__(self, exponent):
        if not isinstance(exponent, Scalar):
            exponent = Scalar(exponent)

        return Scalar(
            value=self.value**exponent.value,
            prevs=[
                Edge(self, exponent.value * self.value ** (exponent.value - 1)),
                Edge(
                    exponent,
                    math.log(abs(self.value)) * (self.value**exponent.value),
                ),
            ],
        )

    @classmethod
    def from_random(cls, lower=0, upper=1):
        return Scalar(random.uniform(lower, upper))
